"""The art module contains all the available arts for your game. You should look to arts based on geometries or on files."""
from .art import Art
from .file import ImageFile, ImageFolder, GIFFile
from .geometry import Polygon, Rectangle, RoundedRectangle, Circle, Ellipse, TexturedCircle, TexturedEllipse, TexturedPolygon, TexturedRoundedRectangle
